/*
 * This file is part of the YASMIN 8051 simulator project
 *
 * Copyright (c) 2000 by Gwenn Englebienne
 *
 * This software is free software; it is protected by the GPL. You can 
 * modify and/or redistribute it under the terms specified by the GNU
 * Public Licence version 2.
 *
 * Be advised that this software comes with no warranty at all, not
 * even the implied warranty of merchantability or fitness for a 
 * particular purpose. 
 *
 * For more information, you could look at the COPYING file, or have a 
 * look at the free software foundation's web site: http://www.fsf.org
 *
 */
 
fr_info.contains(dest_name))
    dest_addr = sfr_info[dest_name].address;
  else
    error ("Unkown operand.");

  if (direct)
    da (dest_addr)--;
  else
    ida (dest_addr)--;
  line_nr++;
}

void digital_adjust (void)
{
  if (strcmp (dest_name, "A"))
    error ("operand of decimal adjust must be accumulator (A)");
  else {
    unsigned char dest_addr = sfr_info["A"].address;
    if ( da (sfr_info["PSW"].address) & CY ||
	 ((da (dest_addr) >> 4) & 0x0F) > 9) {
      if ( (((da (dest_addr) & 0xF0) >> 4) + 0x06) & 0xF0)
	da (sfr_info["PSW"].address) |= CY;
      else
	da (sfr_info["PSW"].address) &= ~CY;
      da (dest_addr) += 0x60;
    }
    if ( da (sfr_info["PSW"].address) & AC || (da (dest_addr) & 0x0F) > 9)
      da (dest_addr) += 0x06;
  }
  line_nr++;
}

void setb (void)
{
  if (sfr_info.contains (dest_name))
    if (sfr_info[dest_name].type == bit)
      set_bit (sfr_info[dest_name].address);
    else
      error ("Not a bit address");
  else if (is_a_nr(dest_name))
    set_bit (asm_to_char(dest_name));
  else
    error ("Only symbolic names and addresses  are supported for now!");
  line_nr++;
}

void clrb (void)
{
  if (sfr_info.contains (dest_name))
    if (sfr_info[dest_name].type == bit)
      clr_bit (sfr_info[dest_name].address);
    else if (!strcmp(dest_name, "A"))
      da (sfr_info["A"].address) = 0x00;
    else
      error ("Not a bit address");
  else if (is_a_nr(dest_name))
    clr_bit (asm_to_char(dest_name));
  else
    error ("Only symbolic names and addresses  are supported for now!");
  line_nr++;
}

void anl (void)
{
  unsigned char dest_addr;
  char source_value;
  int source_is_bit = 0, dest_is_bit = 0;
  if (is_a_nr (dest_name))
    dest_addr = asm_to_char(dest_name);
  else if (!strcmp(dest_name, "A"))
    dest_addr = sfr_info["A"].address;
  else if (sfr_info.contains (dest_name) && sfr_info[dest_name].type == byte)
    dest_addr = sfr_info[dest_name].address;
  else if (sfr_info.contains(dest_name) && sfr_info.get(dest_name).type == bit)
    {
      dest_addr = sfr_info[dest_name].address;
      dest_is_bit = 1;
    }
  else
    error ("Not a good destination address for ANL");
  
  if (is_a_nr(source_name))
    source_value = da(asm_to_char(source_name));
  else if (source_name[0] == '#')
    source_value = asm_to_char (source_name + 1);
  else if (source_name[0] == '@') {
    if (source_name[1] == 'R' && (source_name[2] == '0' ||
				  source_name[2] == '1') &&
	source_name[3] == 0) {
      source_value = ida ( da (sfr_info[source_name +1 ].address));
    }
    else
      error ("Indirect works only with R0 & R1");
  }
  else if (sfr_info.contains(source_name))
    if (sfr_info[source_name].type == bit) {
      source_is_bit = 1;
      source_value = is_bit_set (sfr_info[source_name].address);
      if (DEBUG)
	cerr << "source-value == " << (int)source_value
	     << ", address = " << setbase(16) << (int)sfr_info[source_name].address << setbase(10) << endl;
    }
    else
      source_value = da (sfr_info[source_name].address);
  else
    error ("Invalid source");

  if (DEBUG)
    cerr << "source is bit: " << (source_is_bit ? "true" : "false")
	 << ", val = " << (int)source_value << "\n dest is bit : "
	 << (dest_is_bit ? "true" : "false") << " dest_address = "
	 << (int)dest_addr << " val = " << is_bit_set (dest_addr) << endl;
  
  if (source_is_bit && dest_is_bit)
    if (is_bit_set(dest_addr) && source_value)
      set_bit (dest_addr);
    else
      clr_bit (dest_addr);
  else if (!source_is_bit && !dest_is_bit)
    da (dest_addr) &= source_value;
  else
    error ("Both operands should be of the same type!");
  line_nr++;
}

void orl (void)
{
  unsigned char dest_addr;
  char source_value;
  int source_is_bit = 0, dest_is_bit = 0;
  if (is_a_nr (dest_name))
    dest_addr = asm_to_char(dest_name);
  else if (!strcmp(dest_name, "A"))
    dest_addr = sfr_info["A"].address;
  else if (sfr_info.contains (dest_name) && sfr_info[dest_name].type == byte)
    dest_addr = sfr_info[dest_name].address;
  else if (sfr_info.contains(dest_name) && sfr_info.get(dest_name).type == bit)
    {
      dest_addr = sfr_info[dest_name].address;
      dest_is_bit = 1;
    }
  else
    error ("Not a good destination address for ANL");
  
  if (is_a_nr(source_name))
    source_value = da(asm_to_char(source_name));
  else if (source_name[0] == '#')
    source_value = asm_to_char (source_name + 1);
  else if (source_name[0] == '@') {
    if (source_name[1] == 'R' && (source_name[2] == '0' ||
				  source_name[2] == '1') &&
	source_name[3] == 0) {
      source_value = ida ( da (sfr_info[source_name +1 ].address));
    }
    else
      error ("Indirect works only with R0 & R1");
  }
  else if (sfr_info.contains(source_name))
    if (sfr_info[source_name].type == bit) {
      source_is_bit = 1;
      source_value = is_bit_set (sfr_info[source_name].address);
      if (DEBUG)
	cerr << "source-value == " << (int)source_value
	     << ", address = " << setbase(16) << (int)sfr_info[source_name].address << setbase(10) << endl;
    }
    else
      source_value = da (sfr_info[source_name].address);
  else
    error ("Invalid source");

  if (DEBUG)
    cerr << "source is bit: " << (source_is_bit ? "true" : "false")
	 << ", val = " << (int)source_value << "\n dest is bit : "
	 << (dest_is_bit ? "true" : "false") << " dest_address = "
	 << (int)dest_addr << " val = " << is_bit_set (dest_addr) << endl;
  
  if (source_is_bit && dest_is_bit)
    if (is_bit_set(dest_addr) || source_value)
      set_bit (dest_addr);
    else
      clr_bit (dest_addr);
  else if (!source_is_bit && !dest_is_bit)
    da (dest_addr) |= source_value;
  else
    error ("Both operands should be of the same type!");
  line_nr++;
}

void xorl (void)
{
  unsigned char dest_addr;
  char source_value;
  int source_is_bit = 0, dest_is_bit = 0;
  if (is_a_nr (dest_name))
    dest_addr = asm_to_char(dest_name);
  else if (!strcmp(dest_name, "A"))
    dest_addr = sfr_info["A"].address;
  else if (sfr_info.contains (dest_name) && sfr_info[dest_name].type == byte)
    dest_addr = sfr_info[dest_name].address;
  else if (sfr_info.contains(dest_name) && sfr_info.get(dest_name).type == bit)
    {
      dest_addr = sfr_info[dest_name].address;
      dest_is_bit = 1;
    }
  else
    error ("Not a good destination address for XRL");
  
  if (is_a_nr(source_name))
    source_value = da(asm_to_char(source_name));
  else if (source_name[0] == '#')
    source_value = asm_to_char (source_name + 1);
  else if (source_name[0] == '@') {
    if (source_name[1] == 'R' && (source_name[2] == '0' ||
				  source_name[2] == '1') &&
	source_name[3] == 0) {
      source_value = ida ( da (sfr_info[source_name +1 ].address));
    }
    else
      error ("Indirect works only with R0 & R1");
  }
  else if (sfr_info.contains(source_name))
    if (sfr_info[source_name].type == bit) {
      source_is_bit = 1;
      source_value = is_bit_set (sfr_info[source_name].address);
      if (DEBUG)
	cerr << "source-value == " << (int)source_value
	     << ", address = " << setbase(16) << (int)sfr_info[source_name].address << setbase(10) << endl;
    }
    else
      source_value = da (sfr_info[source_name].address);
  else
    error ("Invalid source");

  if (DEBUG)
    cerr << "source is bit: " << (source_is_bit ? "true" : "false")
	 << ", val = " << (int)source_value << "\n dest is bit : "
	 << (dest_is_bit ? "true" : "false") << " dest_address = "
	 << (int)dest_addr << " val = " << is_bit_set (dest_addr) << endl;
  
  if (source_is_bit && dest_is_bit)
    if (is_bit_set(dest_addr) ^ source_value)
      set_bit (dest_addr);
    else
      clr_bit (dest_addr);
  else if (!source_is_bit && !dest_is_bit)
    da (dest_addr) ^= source_value;
  else
    error ("Both operands should be of the same type!");
  line_nr++;
}

void cpl (void)
{
  if (sfr_info.contains(dest_name))
    if (sfr_info[dest_name].type == bit)
      cpl_bit (sfr_info[dest_name].address);
    else if (!strcmp(dest_name, "A"))
      da (sfr_info["A"].address) = ~da (sfr_info["A"].address);
    else
      error("Wrong operand of complement");
  line_nr++;
}
	     
void push (void)
{
  if (sfr_info.contains(dest_name) && sfr_info[dest_name].type == byte)
    push_on_stack (da (sfr_info[dest_name].address));
  else if (is_a_nr (dest_name))
    push_on_stack ( asm_to_char (dest_name));
  else
    error ("not a valid operand for push instruction");
  line_nr++;
}

void pop (void)
{
  if (sfr_info.contains(dest_name) && sfr_info[dest_name].type == byte)
    da (sfr_info[dest_name].address) = pop_from_stack ();
  else if (is_a_nr (dest_name))
    da (asm_to_char (dest_name)) = pop_from_stack();
  else
    error ("not a valid operand for push instruction");
  line_nr++;
}    

void multiply (void)
{
  if (strcmp(dest_name,"A") || strcmp (source_name, "B"))
    error("Only possible use of mul: MUL A B");
  else {
    int tmp = da (sfr_info["A"].address) * da (sfr_info["B"].address);
    da (sfr_info["A"].address) = (char) (tmp & 0x00FF);
    da (sfr_info["B"].address) = (char) ((tmp & 0x0000FF00) >> 8);
    if (da(sfr_info["B"].address) != 0x00)
      da (sfr_info["PSW"].address) |= OV;
    else
      da (sfr_info["PSW"].address) &= ~OV;    
    clr_bit(sfr_info["C"].address);    
  }
  line_nr++;
}

void divide (void)
{
  char res, mod;
  if (strcmp (dest_name, "A") || strcmp (source_name, "B")) 
    error ("Only A and B can be operands of DIV");
  else {
    if (da (sfr_info["B"].address) == 0)
      da (sfr_info["PSW"].address) |= OV;
    else {
      res = da (sfr_info["A"].address) / da (sfr_info["B"].address);
      mod = da (sfr_info["A"].address) % da (sfr_info["B"].address);
      da (sfr_info["A"].address) = res;
      da (sfr_info["B"].address) = mod;
      da (sfr_info["PSW"].address
	  ) &= ~OV;
    }
  }
  clr_bit (sfr_info["C"].address);
  line_nr++;
}

void jump_on_carry ()
{
  if (is_bit_set (sfr_info["C"].address))
    if (labels.contains(dest_name))
      line_nr = labels[dest_name];
    else
      error("unknown label!");
  else
    line_nr++;
}

void jump_no_carry ()
{
  if (!is_bit_set (sfr_info["C"].address))
    if (labels.contains(dest_name))
      line_nr = labels[dest_name];
    else
      error("unknown label!");
  else
    line_nr++;
}

void jump_on_bit (void)
{
  unsigned char dest_addr;
  unsigned short rel;
  int ok = 1;
  if (is_a_nr (dest_name))
    dest_addr = asm_to_char (dest_name);
  else if (sfr_info.contains(dest_name) && sfr_info[dest_name].type == bit)
    dest_addr = sfr_info[dest_name].address;
  else
    error ("wrong operand #1"), ok = 0;

  if (is_a_nr (source_name))
    rel = asm_to_char(source_name);
  else if (labels.contains(source_name))
    rel = labels[source_name];
  else
    error ("unknown jump address!"), ok = 0;

  if (ok) 
    if (is_bit_set(dest_addr))
      line_nr = rel;
    else
      line_nr++;
  else
    line_nr++;
}   

void jump_no_bit (void)
{
  unsigned char dest_addr;
  unsigned short rel;
  int ok = 1;
  if (is_a_nr (dest_name))
    dest_addr = asm_to_char (dest_name);
  else if (sfr_info.contains(dest_name) && sfr_info[dest_name].type == bit)
    dest_addr = sfr_info[dest_name].address;
  else
    error ("wrong operand #1"), ok = 0;

  if (is_a_nr (source_name))
    rel = asm_to_char(source_name);
  else if (labels.contains(source_name))
    rel = labels[source_name];
  else
    error ("unknown jump address!"), ok = 0;

  if (ok) 
    if (!is_bit_set(dest_addr))
      line_nr = rel;
    else
      line_nr++;
  else
    line_nr++;
}   

void jump_on_bit_and_clear (void)
{
  unsigned char dest_addr;
  unsigned short rel;
  int ok = 1;
  if (is_a_nr (dest_name))
    dest_addr = asm_to_char (dest_name);
  else if (sfr_info.contains(dest_name) && sfr_info[dest_name].type == bit)
    dest_addr = sfr_info[dest_name].address;
  else
    error ("wrong operand #1"), ok = 0;

  if (is_a_nr (source_name))
    rel = asm_to_char(source_name);
  else if (labels.contains(source_name))
    rel = labels[source_name];
  else
    error ("unknown jump address!"), ok = 0;

  if (ok) 
    if (is_bit_set(dest_addr)) {
      clr_bit(dest_addr);
      line_nr = rel;
    }
    else
      line_nr++;
  else
    line_nr++;
}   

void nop (void) {
  line_nr++;
}

void rl (void)
{
  if (strcmp(dest_name, "A"))
    error ("Only A can be rotated!");
  else {
    int tmp = da (sfr_info["A"].address);
    int last_bit = (da (sfr_info["A"].address) & 0x80) != 0;
    tmp <<= 1;
    if (last_bit)
      tmp |= 0x01;
    else
      tmp &= 0xFE;
    da (sfr_info["A"].address) = tmp;
  }
  line_nr++;
}

void rlc (void)
{
  if (strcmp(dest_name, "A"))
    error ("Only A can be rotated!");
  else {
    int tmp = da (sfr_info["A"].address);
    int last_bit = is_bit_set(sfr_info["C"].address);
    if ((da (sfr_info["A"].address) & 0x80) != 0)
      set_bit(sfr_info["C"].address);
    else
      clr_bit(sfr_info["C"].address);
    tmp <<= 1;
    if (last_bit)
      tmp |= 0x01;
    else
      tmp &= 0xFE;
    da (sfr_info["A"].address) = tmp;
  }
  line_nr++;
}

void rr (void)
{
  if (strcmp(dest_name, "A"))
    error ("Only A can be rotated!");
  else {
    int tmp = da (sfr_info["A"].address);
    int last_bit = (da (sfr_info["A"].address) & 0x01) != 0;
    tmp >>= 1;
    if (last_bit)
      tmp |= 0x80;
    else
      tmp &= 0x7F;
    da (sfr_info["A"].address) = tmp;
  }
  line_nr++;
}  

void rrc (void)
{
    if (strcmp(dest_name, "A"))
    error ("Only A can be rotated!");
  else {
    int tmp = da (sfr_info["A"].address);
    int last_bit = is_bit_set(sfr_info["C"].address);
    if ((da (sfr_info["A"].address) & 0x01) != 0)
      set_bit (sfr_info["C"].address);
    else
      clr_bit (sfr_info["C"].address);
    tmp >>= 1;
    if (last_bit)
      tmp |= 0x80;
    else
      tmp &= 0x7F;
    da (sfr_info["A"].address) = tmp;
  }
  line_nr++;
}  

void swap (void)
{
  if (strcmp(dest_name,"A"))
    error ("Only Accumulator can be swapped");
  else {
    unsigned char high, low;
    high = da (sfr_info["A"].address);
    low = (high & 0x0F) << 4;
    high = ((high & 0xF0) >> 4) & 0x0f; // to avoid problems
    da (sfr_info["A"].address) = high | low;
  }
  line_nr++;
}

void exchange (void)
{
     if (strcmp("A",dest_name))
	  error ("XCH can only have the Acc as destination");
     else {
	  char tmp = da (sfr_info["A"].address);
	  char* source_address;
	  if (is_a_nr(source_name))
	       source_address = & da (asm_to_char(source_name));
	  else if (source_name[0] == '#')
	       error ("You can't XCH A with data!");
	  else if (source_name[0] == '@') {
	       if (source_name[1] == 'R' && (source_name[2] == '0' ||
					     source_name[2] == '1') &&
		   source_name[3] == 0) {
		    source_address = & ida ( da (sfr_info[source_name +1 ].address));
	       }
	       else
		    error ("Indirect works only with R0 & R1");
	  }
	  else if (sfr_info.contains(source_name)) 
	       source_address = & da (sfr_info[source_name].address);
	  else
	       error ("Invalid source");
	  da (sfr_info["A"].address) = *source_address;
	  *source_address = tmp;
     }
     line_nr++;
}

void exchange_nibble (void)
{
     if (strcmp("A", dest_name))
	  error ("XCHD only works with Acc as destination.");
     else {
	  char tmp = da (sfr_info["A"].address);
	  char* source_address;
	  if (source_name[0] == '@') {
	       if (source_name[1] == 'R' && (source_name[2] == '0' ||
					     source_name[2] == '1') &&
		   source_name[3] == 0) {
		    source_address = & ida ( da (sfr_info[source_name +1 ].address));
	       }
	       else
		    error ("Indirect works only with R0 & R1");
	  }
	  else
	       error ("Invalid source");
	  da (sfr_info["A"].address) &= 0x0F;
	  da (sfr_info["A"].address) |= (*source_address & 0xF0);
	  *source_address &= 0x0f;
	  *source_address |= tmp & 0xF0;
     }
     line_nr++;
}	  

void cmp_jump_on_not_equal (void)
{
     char dest, source;
     if (!strcmp("A", dest_name) ||
	 !strcmp("Acc", dest_name))
	  dest = da(sfr_info["A"].address);
     else if (dest_name[0] == 'R' &&
	      (dest_name[1] >= '0' && dest_name[1] <= '7') &&
	      dest_name[2] == 0)
	  dest = da (sfr_info[dest_name].address);
     else if (dest_name[0] == '@' &&
	      dest_name[1] == 'R' &&
	      (dest_name[2] == '1' || dest_name[2] == '0') &&
	      dest_name[3] == 0)
	  dest = ida ( da (sfr_info[dest_name + 1].address) );
     else
	  error ("CJNE Wrong destination parameter");

     if (is_a_nr (source_name))
	  source = da (asm_to_char(source_name));
     else if (source_name[0] == '#' && is_a_nr (source_name + 1))
	  source = asm_to_char (source_name + 1);
     else
	  error ("CJNE Bad source parameter");

     if (source != dest)
	  if (labels.contains(op3_name))
	       line_nr = labels[op3_name];
	  else
	       error ("Unknown label!");
     else
	  line_nr++;

     if (DEBUG)
	  cout << "CJNE called. dest = " << (int)dest << ", source = "
	       << (int)source << ", label = " << op3_name
	       << " of which line nr = " << labels.get (op3_name) << endl;
	       
}

void create_function_table (void)
{
  if (DEBUG)
    cerr << "Creating instructions table... ";
  instructions["ADD"] = add; 
  instructions["ADDC"] = addc;
  instructions["SUBB"] = sub_borrow;
  instructions["INC"] = inc;  
  instructions["DEC"] = dec;
  instructions["MUL"] = multiply;  
  instructions["DIV"] = divide;
  instructions["DA"] = digital_adjust;  
  instructions["ANL"] = anl;
  instructions["ORL"] = orl;  
  instructions["XRL"] = xorl;
  instructions["CLR"] = clrb;  
  instructions["CPL"] = cpl;
  instructions["RL"] = rl;  
  instructions["RLC"] = rlc;
  instructions["RR"] = rr;  
  instructions["RRC"] = rrc;
  instructions["SWAP"] = swap;  
  instructions["MOV"] = move;
  instructions["MOVX"] = unknown;  
  instructions["PUSH"] = push;
  instructions["POP"] = pop;  
  instructions["XCH"] = exchange;
  instructions["XCHD"] = exchange_nibble;  
  instructions["SETB"] = setb;  
  instructions["ACALL"] = call;
  instructions["LCALL"] = call;
  instructions["CALL"] = call;
  instructions["RET"] = ret;
  instructions["RETI"] = reti;  
  instructions["AJMP"] = jump;          
  instructions["LJMP"] = jump;          
  instructions["SJMP"] = jump;
  instructions["JMP"] = jump;           
  instructions["JZ"] = jump_on_zero;
  instructions["JNZ"] = jump_not_zero;  
  instructions["JC"] = jump_on_carry;
  instructions["JNC"] = jump_no_carry;
  instructions["JB"] = jump_on_bit;
  instructions["JNB"] = jump_no_bit;  
  instructions["JBC"] = jump_on_bit_and_clear;
  instructions["CJNE"] = cmp_jump_on_not_equal;  
  instructions["DJNZ"] = djnz;
  instructions["NOP"] = nop;
  instructions[""] = nop; // for labels on an empty line

  instructions["MOVC"] = no_movc;
  
  if (DEBUG)
    cerr << "Done.\n";
}
